# Bunch of declarations from C to python. The idea here is to define
# only the quantities that will be used, for input, output or
# intermediate manipulation, by the python wrapper. For instance, in
# the precision structure, the only item used here are the error
# message and one parameter used for an error message. That is why
# nothing more is defined from this structure. The rest is internal in
# Class.  If, for whatever reason, you need an other, existing
# parameter from Class, remember to add it inside this cdef.

cdef extern from "class.h":

    cdef char[10] _VERSION_

    ctypedef char FileArg[40]

    ctypedef char* ErrorMsg

    ctypedef char FileName[256]

    cdef enum interpolation_method:
        inter_normal
        inter_growing_closeby

    cdef enum vecback_format:
        short_info
        normal_info
        long_info

    cdef enum linear_or_logarithmic:
        linear
        logarithmic

    cdef enum file_format:
        class_format
        camb_format

    cdef enum non_linear_method:
        nl_none
        nl_halofit
        nl_HMcode

    cdef enum pk_outputs:
        pk_linear
        pk_nonlinear
        pk_numerical_nowiggle
        pk_analytic_nowiggle

    cdef enum out_sigmas:
        out_sigma
        out_sigma_prime
        out_sigma_disp

    cdef struct precision:
        double nonlinear_min_k_max
        ErrorMsg error_message

    cdef struct background:
        short is_allocated
        ErrorMsg error_message
        int bg_size
        int index_bg_ang_distance
        int index_bg_lum_distance
        int index_bg_conf_distance
        int index_bg_a
        int index_bg_H
        int index_bg_D
        int index_bg_f
        int index_bg_Omega_m
        int index_bg_rho_b
        int index_bg_rho_cdm
        int index_bg_rho_ncdm1
        int index_bg_rho_crit
        short has_cdm
        short  has_ncdm
        int N_ncdm
        double T_cmb
        double h
        double H0
        double age
        double conformal_age
        double K
        double * m_ncdm_in_eV
        double Neff
        double Omega0_g
        double Omega0_b
        double Omega0_idr
        double T_idr
        double Omega0_cdm
        double Omega0_idm
        double Omega0_dcdm
        double Omega0_ncdm_tot
        double Omega0_lambda
        double Omega0_fld
        double w0_fld
        double wa_fld
        double cs2_fld
        double Omega0_ur
        double Omega0_dcdmdr
        double Omega0_dr
        double Omega0_scf
        double Omega0_k
        int bt_size
        double Omega0_m
        double Omega0_r
        double Omega0_de
        double a_eq
        double H_eq
        double z_eq
        double tau_eq

    cdef struct thermodynamics:
        short is_allocated
        ErrorMsg error_message
        int th_size
        int index_th_xe
        int index_th_Tb
        double tau_reio
        double z_reio
        double z_rec
        double tau_rec
        double rs_rec
        double ds_rec
        double da_rec
        double z_star
        double tau_star
        double rs_star
        double ds_star
        double ra_star
        double da_star
        double rd_star
        double z_d
        double tau_d
        double ds_d
        double rs_d
        double conf_time_reio
        double YHe
        double n_e
        double a_idm_dr
        double b_idr
        double nindex_idm_dr
        double m_idm
        double cross_idm_g
        double u_idm_g
        double cross_idm_b
        double n_index_idm_b
        int tt_size

    cdef struct perturbations:
        short is_allocated
        ErrorMsg error_message
        short has_scalars
        short has_vectors
        short has_tensors

        short has_density_transfers
        short has_velocity_transfers

        int has_pk_matter
        int l_lss_max

        int store_perturbations
        int k_output_values_num
        double k_output_values[30]
        double k_max_for_pk
        int index_k_output_values[30]
        char scalar_titles[8000]
        char vector_titles[8000]
        char tensor_titles[8000]
        int number_of_scalar_titles
        int number_of_vector_titles
        int number_of_tensor_titles
        int index_md_scalars

        double * scalar_perturbations_data[30]
        double * vector_perturbations_data[30]
        double * tensor_perturbations_data[30]
        int size_scalar_perturbation_data[30]
        int size_vector_perturbation_data[30]
        int size_tensor_perturbation_data[30]

        double * alpha_idm_dr
        double * beta_idr

        # add source functions for comparison
        short has_source_t
        short has_source_p
        short has_source_delta_m
        short has_source_delta_cb
        short has_source_delta_tot
        short has_source_delta_g
        short has_source_delta_b
        short has_source_delta_cdm
        short has_source_delta_idm
        short has_source_delta_idr
        short has_source_delta_dcdm
        short has_source_delta_fld
        short has_source_delta_scf
        short has_source_delta_dr
        short has_source_delta_ur
        short has_source_delta_ncdm
        short has_source_theta_m
        short has_source_theta_cb
        short has_source_theta_tot
        short has_source_theta_g
        short has_source_theta_b
        short has_source_theta_cdm
        short has_source_theta_idm
        short has_source_theta_idr
        short has_source_theta_dcdm
        short has_source_theta_fld
        short has_source_theta_scf
        short has_source_theta_dr
        short has_source_theta_ur
        short has_source_theta_ncdm
        short has_source_phi
        short has_source_phi_prime
        short has_source_phi_plus_psi
        short has_source_psi
        short has_source_h
        short has_source_h_prime
        short has_source_eta
        short has_source_eta_prime
        short has_source_H_T_Nb_prime
        short has_source_k2gamma_Nb

        int index_tp_t0
        int index_tp_t1
        int index_tp_t2
        int index_tp_p
        int index_tp_delta_m
        int index_tp_delta_cb
        int index_tp_delta_tot
        int index_tp_delta_g
        int index_tp_delta_b
        int index_tp_delta_cdm
        int index_tp_delta_idm
        int index_tp_delta_dcdm
        int index_tp_delta_fld
        int index_tp_delta_scf
        int index_tp_delta_dr
        int index_tp_delta_ur
        int index_tp_delta_idr
        int index_tp_delta_ncdm1

        int index_tp_theta_m
        int index_tp_theta_cb
        int index_tp_theta_tot
        int index_tp_theta_g
        int index_tp_theta_b
        int index_tp_theta_cdm
        int index_tp_theta_dcdm
        int index_tp_theta_fld
        int index_tp_theta_scf
        int index_tp_theta_ur
        int index_tp_theta_idr
        int index_tp_theta_idm
        int index_tp_theta_dr
        int index_tp_theta_ncdm1

        int index_tp_phi
        int index_tp_phi_prime
        int index_tp_phi_plus_psi
        int index_tp_psi
        int index_tp_h
        int index_tp_h_prime
        int index_tp_eta
        int index_tp_eta_prime
        int index_tp_H_T_Nb_prime
        int index_tp_k2gamma_Nb


        double *** sources
        double * tau_sampling
        int tau_size
        int k_size_pk
        int * k_size
        double ** k
        int * ic_size
        int index_ic_ad
        int md_size
        int * tp_size
        double * ln_tau
        int ln_tau_size

    cdef struct transfer:
        short is_allocated
        ErrorMsg error_message

    cdef struct primordial:
        short is_allocated
        ErrorMsg error_message
        double k_pivot
        double A_s
        double n_s
        double alpha_s
        double beta_s
        double r
        double n_t
        double alpha_t
        double V0
        double V1
        double V2
        double V3
        double V4
        double f_cdi
        double n_cdi
        double c_ad_cdi
        double n_ad_cdi
        double f_nid
        double n_nid
        double c_ad_nid
        double n_ad_nid
        double f_niv
        double n_niv
        double c_ad_niv
        double n_ad_niv
        double phi_min
        double phi_max
        int lnk_size

    cdef struct harmonic:
        short is_allocated
        ErrorMsg error_message
        int has_tt
        int has_te
        int has_ee
        int has_bb
        int has_pp
        int has_tp
        int has_dd
        int has_td
        int has_ll
        int has_dl
        int has_tl
        int l_max_tot
        int ** l_max_ct
        int ct_size
        int * ic_size
        int * ic_ic_size
        int md_size
        int d_size
        int non_diag
        int index_ct_tt
        int index_ct_te
        int index_ct_ee
        int index_ct_bb
        int index_ct_pp
        int index_ct_tp
        int index_ct_dd
        int index_ct_td
        int index_ct_pd
        int index_ct_ll
        int index_ct_dl
        int index_ct_tl
        int * l_size
        int index_md_scalars

    cdef struct output:
        ErrorMsg error_message

    cdef struct distortions:
        short is_allocated
        double * sd_parameter_table
        int index_type_g
        int index_type_mu
        int index_type_y
        int index_type_PCA
        int type_size
        double * DI
        double * x
        double DI_units
        double x_to_nu
        int has_distortions
        int x_size
        ErrorMsg error_message

    cdef struct lensing:
        short is_allocated
        int has_tt
        int has_ee
        int has_te
        int has_bb
        int has_pp
        int has_tp
        int has_dd
        int has_td
        int has_ll
        int has_dl
        int has_tl
        int index_lt_tt
        int index_lt_te
        int index_lt_ee
        int index_lt_bb
        int index_lt_pp
        int index_lt_tp
        int index_lt_dd
        int index_lt_td
        int index_lt_ll
        int index_lt_dl
        int index_lt_tl
        int * l_max_lt
        int lt_size
        int has_lensed_cls
        int l_lensed_max
        int l_unlensed_max
        ErrorMsg error_message

    cdef struct fourier:
        short is_allocated
        short has_pk_matter
        short has_pk_numerical_nowiggle
        short has_pk_analytic_nowiggle
        int method
        int ic_size
        int ic_ic_size
        int k_size
        int k_size_pk
        int ln_tau_size
        int index_ln_tau_pk
        int tau_size
        int index_tau_min_nl
        double * k
        double * ln_tau
        double * tau
        double ** ln_pk_l
        double ** ln_pk_nl
        double * sigma8
        int has_pk_m
        int has_pk_cb
        int index_pk_m
        int index_pk_cb
        int index_pk_total
        int index_pk_cluster
        ErrorMsg error_message

    cdef struct file_content:
        char * filename
        int size
        FileArg * name
        FileArg * value
        short * read

    void lensing_free(void*)
    void harmonic_free(void*)
    void transfer_free(void*)
    void primordial_free(void*)
    void perturbations_free(void*)
    void thermodynamics_free(void*)
    void background_free(void*)
    void fourier_free(void*)
    void distortions_free(void*)

    cdef int _FAILURE_
    cdef int _FALSE_
    cdef int _TRUE_

    cdef double _Mpc_over_m_
    cdef double _c_
    cdef double _G_
    cdef double _eV_

    int input_read_from_file(void*, void*, void*, void*, void*, void*, void*, void*, void*,
        void*, void*, void*, char*)
    int background_init(void*,void*)
    int thermodynamics_init(void*,void*,void*)
    int perturbations_init(void*,void*,void*,void*)
    int primordial_init(void*,void*,void*)
    int fourier_init(void*,void*,void*,void*,void*,void*)
    int transfer_init(void*,void*,void*,void*,void*,void*)
    int harmonic_init(void*,void*,void*,void*,void*,void*,void*)
    int lensing_init(void*,void*,void*,void*,void*)
    int distortions_init(void*,void*,void*,void*,void*,void*)

    int background_tau_of_z(void* pba, double z,double* tau)
    int background_z_of_tau(void* pba, double tau,double* z)
    int background_at_z(void* pba, double z, int return_format, int inter_mode, int * last_index, double *pvecback)
    int background_at_tau(void* pba, double tau, int return_format, int inter_mode, int * last_index, double *pvecback)
    int background_output_titles(void * pba, char titles[8000])
    int background_output_data(void *pba, int number_of_titles, double *data)

    int thermodynamics_at_z(void * pba, void * pth, double z, int inter_mode, int * last_index, double *pvecback, double *pvecthermo)
    int thermodynamics_output_titles(void * pba, void *pth, char titles[8000])
    int thermodynamics_output_data(void *pba, void *pth, int number_of_titles, double *data)

    int perturbations_output_data_at_z(void *pba,void *ppt, file_format output_format, double z, int number_of_titles, double *data)
    int perturbations_output_data_at_index_tau(void *pba,void *ppt, file_format output_format, int ondex_tau, int number_of_titles, double *data)
    int perturbations_output_data(void *pba,void *ppt, file_format output_format, double * tkfull, int number_of_titles, double *data)
    int perturbations_output_firstline_and_ic_suffix(void *ppt, int index_ic, char first_line[1024], FileName ic_suffix)
    int perturbations_output_titles(void *pba, void *ppt,  file_format output_format, char titles[8000])

    int primordial_output_titles(void * ppt, void *ppm, char titles[8000])
    int primordial_output_data(void *ppt, void *ppm, int number_of_titles, double *data)

    int harmonic_cl_at_l(void* phr,double l,double * cl,double * * cl_md,double * * cl_md_ic)
    int lensing_cl_at_l(void * ple,int l,double * cl_lensed)

    int harmonic_pk_at_z(
        void * pba,
        void * phr,
        int mode,
        double z,
        double * output_tot,
        double * output_ic,
        double * output_cb_tot,
        double * output_cb_ic
        )
    int fourier_pk_at_z(
        void * pba,
        void *pfo,
        int mode,
        int pk_output,
        double z,
        int index_pk,
        double * out_pk,
        double * out_pk_ic
        )

    int harmonic_pk_at_k_and_z(
        void* pba,
        void * ppm,
        void * phr,
        double k,
        double z,
        double * pk,
        double * pk_ic,
        double * pk_cb,
        double * pk_cb_ic)

    int harmonic_pk_nl_at_k_and_z(
        void* pba,
        void * ppm,
        void * phr,
        double k,
        double z,
        double * pk,
        double * pk_cb)

    int harmonic_pk_nl_at_z(
        void * pba,
        void * phr,
        int mode,
        double z,
        double * output_tot,
        double * output_cb_tot)

    int fourier_pk_at_k_and_z(
        void * pba,
        void * ppm,
        void * pfo,
        int pk_output,
        double k,
        double z,
        int index_pk,
        double * out_pk,
        double * out_pk_ic)

    int fourier_pk_tilt_at_k_and_z(
        void * pba,
        void * ppm,
        void * pfo,
        int pk_output,
        double k,
        double z,
        int index_pk,
        double * pk_tilt)

    int fourier_sigmas_at_z(
        void * ppr,
        void * pba,
        void * pfo,
        double R,
        double z,
        int index_pk,
        int sigma_output,
        double * result)

    int fourier_pks_at_kvec_and_zvec(
        void * pba,
        void * pfo,
        int pk_output,
        double * kvec,
        int kvec_size,
        double * zvec,
        int zvec_size,
        double * out_pk,
        double * out_pk_cb)

    int fourier_k_nl_at_z(void* pba, void* pfo, double z, double* k_nl, double* k_nl_cb)

    int harmonic_firstline_and_ic_suffix(void *ppt, int index_ic, char first_line[1024], FileName ic_suffix)

    int harmonic_fast_pk_at_kvec_and_zvec(
                  void * pba,
                  void * phr,
                  double * kvec,
                  int kvec_size,
                  double * zvec,
                  int zvec_size,
                  double * pk_tot_out,
                  double * pk_cb_tot_out,
                  int nonlinear)



from setuptools import setup
from setuptools import Extension
from Cython.Distutils import build_ext

import numpy as nm
import os
import subprocess as sbp
import os.path as osp

# Recover the gcc compiler
GCCPATH_STRING = sbp.Popen(
    ['gcc', '-print-libgcc-file-name'],
    stdout=sbp.PIPE).communicate()[0]
GCCPATH = osp.normpath(osp.dirname(GCCPATH_STRING)).decode()

liblist = ["class"]
MVEC_STRING = sbp.Popen(
    ['gcc', '-lmvec'],
    stderr=sbp.PIPE).communicate()[1]
if b"mvec" not in MVEC_STRING:
    liblist += ["mvec","m"]

# define absolute paths
root_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
include_folder = os.path.join(root_folder, "include")
classy_folder = os.path.join(root_folder, "python")
heat_folder = os.path.join(os.path.join(root_folder, "external"),"heating")
recfast_folder = os.path.join(os.path.join(root_folder, "external"),"RecfastCLASS")
hyrec_folder = os.path.join(os.path.join(root_folder, "external"),"HyRec2020")
hmcode_folder = os.path.join(os.path.join(root_folder, "external"),"HMcode")
halofit_folder = os.path.join(os.path.join(root_folder, "external"),"Halofit")

# Recover the CLASS version
with open(os.path.join(include_folder, 'common.h'), 'r') as v_file:
    for line in v_file:
        if line.find("_VERSION_") != -1:
            # get rid of the " and the v
            VERSION = line.split()[-1][2:-1]
            break

# Define cython extension and fix Python version
classy_ext = Extension("classy", [os.path.join(classy_folder, "classy.pyx")],
                           include_dirs=[nm.get_include(), include_folder, heat_folder, recfast_folder, hyrec_folder, hmcode_folder, halofit_folder],
                           libraries=liblist,
                           library_dirs=[root_folder, GCCPATH],
                           #extra_link_args=['-lgomp'],
                           language="c++",
                           extra_compile_args=["-std=c++11"]
                       )
import sys
classy_ext.cython_directives = {'language_level': "3" if sys.version_info.major>=3 else "2"}

setup(
    name='classy',
    version=VERSION,
    description='Python interface to the Cosmological Boltzmann code CLASS',
    url='http://www.class-code.net',
    cmdclass={'build_ext': build_ext},
    ext_modules=[classy_ext],
)



"""
Automatically reads header files to generate an interface
"""
from __future__ import division, print_function
import sys
import logging
try:
    from collections import OrderedDict as od
except ImportError:
    try:
        from ordereddict import OrderedDict as od
    except ImportError:
        raise ImportError(
            "If you are running with Python v2.5 or 2.6"
            " you need to manually install the ordereddict"
            " package.")
try:
    import colorlog
except ImportError:
    raise ImportError(
        "You have to install the colorlog module"
        " with pip, or easy-install.")

SPACING = '    '
NAMING_CONVENTION = {
    'precision': {'python': 'precision',
                  'function': 'precision'},
    'background': {'python': 'background',
                   'function': 'background'},
    'thermo': {'python': 'thermodynamics',
               'function': 'thermodynamics'},
    'perturbs': {'python': 'perturbations',
                 'function': 'perturb'},
    'transfers': {'python': 'transfer',
                  'function': 'transfer'},
    'primordial': {'python': 'primordial',
                   'function': 'primordial'},
    'spectra': {'python': 'spectra',
                'function': 'spectra'},
    'lensing': {'python': 'lensing',
                'function': 'lensing'},
    'nonlinear': {'python': 'nonlinear',
                  'function': 'nonlinear'},
    'output': {'python': 'output',
               'function': 'output'},
    }


def main():
    # create logger
    logger = create_logger()

    # Recover all sub-header files
    main_header = '../include/class.h'
    headers = []

    with open(main_header, 'r') as header_file:
        in_modules = False
        for line in header_file:
            if in_modules:
                if line.strip() == '':
                    in_modules = False
                    continue
                if line.find('common') == -1 and line.find('input') == -1:
                    headers.append(
                        '../include/%s' % line.split()[-1].strip('"'))
            if line.find('class modules') != -1:
                in_modules = True

    logger.info('Extracted the following headers: %s', ', '.join(headers))
    output = 'classy.pyx'
    logger.info('Creating %s', output)
    structs = od()
    output_file = open(output, 'w')
    write_imports(output_file)
    output_file.write('cdef extern from "class.h":\n')

    # First write the first non automatic bits
    output_file.write(
        SPACING+'ctypedef char FileArg[40]\n' +
        SPACING+'ctypedef char* ErrorMsg\n' +
        SPACING+'cdef struct precision:\n' +
        2*SPACING+'ErrorMsg error_message\n\n' +
        SPACING+'cdef int _FAILURE_\n' +
        SPACING+'cdef int _FALSE_\n' +
        SPACING+'cdef int _TRUE_\n')

    for header in headers:
        extract_headers(header, structs, output_file, logger)
    logger.info("Finished extracting headers")
    for struct_name, struct in structs.items():
        create_wrapper_class(struct_name, struct, output_file, logger)

    return


def extract_headers(header, structs, output_file, logger):
    """toto"""
    # Initialise the two flags controlling the exploration of the main
    # structure
    in_struct, main_struct_finished = False, False
    # Flags for exploring enums (only the ones before the struct)
    in_enum = False
    # flag dealing with extracting docstrings
    comment_partially_recovered = False
    # Flag keeping track of multiple variables
    multiple_var = False
    # Flag recovering the functions
    in_function_definitions, in_function, in_init = False, False, False
    with open(header, 'r') as header_file:
        logger.info("reading %s" % header)
        for line in header_file:
            # First case, recover the enums
            if not main_struct_finished and not in_struct:
                if line.find("enum ") != -1 and line.find("{") != -1:
                    enum_members = []
                    if line.find(";") == -1:
                        in_enum = True
                        enum_name = line.strip("enum").strip().strip('{')
                    else:
                        in_enum = False
                        line = line.strip("enum").strip().strip(';')
                        enum_name, enum_sign = line.split(' ', 1)
                        enum_sign = enum_sign.strip('}').strip('{')
                        for elem in enum_sign.split(','):
                            enum_members.append(elem.strip())
                        output_file.write(
                            SPACING + 'cdef enum %s:\n' % enum_name)
                        for elem in enum_members:
                            output_file.write(2*SPACING + elem + '\n')
                        output_file.write('\n')
                elif in_enum:
                    if line.find('};') != -1:
                        in_enum = False
                        output_file.write(
                            SPACING + 'cdef enum %s:\n' % enum_name)
                        for elem in enum_members:
                            output_file.write(2*SPACING + elem + '\n')
                        output_file.write('\n')
                    else:
                        if line.strip() != '':
                            enum_members.append(line.split()[0].strip().strip(','))
            if line.find("struct ") != -1 and not main_struct_finished:
                in_struct = True
                # Recover the name
                logger.debug("in struct: %s" % line)
                struct_name = line.strip().split()[1]
                logger.debug("struct name: %s" % struct_name)
                structs[struct_name] = {}
                structs[struct_name].update(
                    NAMING_CONVENTION[struct_name])
                output_file.write("%scdef struct %s:\n" % (
                    SPACING, struct_name))
                continue
            elif in_struct:
                if line.find("};\n") != -1:
                    output_file.write('\n')
                    in_struct, main_struct_finished = False, True
                else:
                    # if the line is not empty or does not contain only a
                    # comment:
                    if line.strip() == '' or line.strip()[:2] == '/*':
                        continue
                    logger.debug(
                        "potentially non empty line: %s" % line.strip())
                    #elif line.find('/**') != -1 or line.find('*/') != -1:
                        #continue
                    if line.find(';') == -1 and not comment_partially_recovered:
                        logger.debug("--> Discarded")
                        continue
                    elif line.find(';') != -1 and not comment_partially_recovered:
                        var_doc = ''
                        var_part, begin_comment = line.strip().split(';', 1)
                        var_doc += begin_comment.strip()[4:].strip()
                        # 2 things can happen: there can be arrays, and there
                        # can be several variables defined in one line...
                        # If array, slightly more complex
                        if var_part.find('*') != -1:
                            # if no comma is found, it means it is a single
                            # variable: good !
                            if var_part.find(',') == -1:
                                # remove if commented (starts with /*)
                                if var_part[:2] in ['/*', '//']:
                                    continue
                                multiple_var = False
                                var_type, var_stars, var_name = var_part.strip().split()
                                structs[struct_name][var_name] = [
                                    var_type, var_stars]
                            else:
                                # Count how many variables are defined
                                multiple_var = True
                                all_vars = [elem.strip() for elem in
                                            var_part.split('*')[-1].split(',')]
                                var_type, var_stars = (var_part.strip().
                                                       split()[:2])
                                for var in all_vars:
                                    structs[struct_name][var] = [
                                        var_type, var_stars]
                        else:
                            # Again check for more than one variable
                            var_stars = ''
                            if var_part.find(',') == -1:
                                multiple_var = False
                                var_type, var_name = var_part.strip().split(' ', 1)
                                # Check if enum
                                if var_type == 'enum':
                                    enum_name, var_name = var_name.split()
                                    var_type += ' '+enum_name
                                structs[struct_name][var_name] = [
                                    var_type, var_stars]
                            else:
                                multiple_var = True
                                all_vars = [elem.strip() for elem in
                                            var_part.split()[2:].split(',')]
                                var_type = (var_part.strip().split()[0])
                                for var in all_vars:
                                    structs[struct_name][var] = [
                                        var_type, var_stars]
                        # If the comment is finished, pass
                        if var_doc[-2:] != '*/':
                            comment_partially_recovered = True
                        else:
                            var_doc = var_doc[:-2].replace('\\f$', '$').strip()
                            structs[struct_name][var_name].append(var_doc)
                            logger.debug(
                                "extracted the variable %s, " % var_name +
                                "of type %s, with docstring: %s" % (
                                    ''.join([var_stars, var_type]), var_doc))

                        if not multiple_var:
                            output_file.write(2*SPACING+' '.join(
                                [elem for elem in [var_type, var_stars, var_name]
                                 if elem])+'\n')
                        else:
                            for var in all_vars:
                                output_file.write(2*SPACING+' '.join(
                                    [elem for elem in [var_type, var_stars, var]
                                     if elem])+'\n')

                    if comment_partially_recovered:
                        logger.debug("--> Accepted")
                        var_doc += ' '+line.strip()
                        if var_doc[-2:] == '*/':
                            comment_partially_recovered = False
                            var_doc = var_doc[:-2].replace('\\f$', '$').strip()
                            structs[struct_name][var_name].append(var_doc)
                            logger.debug(
                                "extracted the variable %s, " % var_name +
                                "of type %s, with docstring: %s" % (
                                    ''.join([var_stars, var_type]), var_doc))

            elif main_struct_finished:
                if line.find('extern "C"') != -1:
                    in_function_definitions = True
                if not in_function_definitions:
                    continue
                else:
                    if line.find('(') != -1:
                        in_function = True
                        logger.debug("Found a function")
                        func_type, func_name = line.split('(')[0].strip().split()
                        logger.debug('%s %s' % (func_name, func_type))
                        func_param = []

                        if func_name == structs[struct_name]['function']+'_init':
                            logger.info("found the init function")
                            in_init = True
                            structs[struct_name]['init'] = [func_name]
                        output_file.write(SPACING+'%s %s(' % (
                            func_type, func_name))
                    elif in_function:
                        # recover the signature of the function
                        line = line.strip().strip(',')
                        if line.find('struct') != -1:
                            if in_init:
                                name = line.split('*')[0].strip()[7:]
                                structs[struct_name]['init'].append(name)
                            func_param.append('void *')
                        elif line.find('*') != -1:
                            # Taking into account with or without spaces
                            temp = ''.join(line.strip(',').split())
                            last_star = len(temp)-temp[::-1].find('*')
                            func_param.append(temp[:last_star])
                        elif line.find(')') == -1:
                            if line != '':
                                func_param.append(line.split()[0])
                        else:
                            logger.debug('signature extracted')
                            in_function = False
                            if in_init:
                                in_init = False
                            output_file.write(', '.join(func_param) + ')\n')

                    elif line.find('}') != -1:
                        output_file.write('\n')
                        in_function_definitions = False
            #print line.strip()


def create_wrapper_class(struct_name, struct, of, logger):
    """TODO"""
    of.write('# Defining wrapper around struct %s\n' % struct_name)
    of.write('cdef class %s:\n' % (
        NAMING_CONVENTION[struct_name]['python'].capitalize()))

    ## recover the number of additional arguments:
    init_name, argument_names = struct['init'][0], struct['init'][1:]

    for companion in argument_names:
        of.write(SPACING+'cdef %s _%s\n' % (companion, companion))
        #logger.info("structure: %s, python name: %s" % (
            #companion, NAMING_CONVENTION[companion]['python']))
    of.write('\n')

    # Define the array variables for all needed
    array_variables = []
    variables = []
    for key, value in struct.items():
        if key != 'init':
            if value[1]:
                array_variables.append(key)
                variables.append(key)
                of.write(SPACING+'cdef np.ndarray %s_arr\n' % key)
            else:
                variables.append(key)
    of.write('\n')

    # write the init
    of.write(SPACING+'def __init__(self')
    for companion in argument_names:
        of.write(", %s py_%s" % (
            NAMING_CONVENTION[companion]['python'].capitalize(), companion))
    of.write('):\n\n')

    # pointing the pointers where they belong
    for companion in argument_names:
        of.write(2*SPACING+"self._%s = py_%s._%s\n" % (
            companion, companion, companion))

    # Writing the call to structname_init()
    of.write(2*SPACING+'%s_init(\n' % struct_name)
    for companion in argument_names:
        of.write(3*SPACING+'&(self._%s),\n' % companion)
    of.write(3*SPACING+'&(self._%s))\n\n' % struct_name)

    #of.write(2*SPACING+'%s_init(&(self._%s))\n\n' % (
        #struct_name, struct_name))
    for array in array_variables:
        of.write(2*SPACING+'# Wrapping %s\n' % array)
        of.write(2*SPACING+'%s_wrapper = ArrayWrapper()\n' % array)
        of.write(
            2*SPACING+"%s_wrapper.set_data(%d, '%s', "
            "<void*> self._%s.%s)\n" % (
                array, 2, struct[array].strip('*'), struct_name, array))
        of.write(
            2*SPACING+'self.%s_arr = np.array(%s_wrapper, '
            'copy=False)\n' % (
                array, array))
        of.write(2*SPACING+'self.%s_arr.base = '
                 '<PyObject*> %s_wrapper\n' % (
                     array, array))
        of.write(2*SPACING+'Py_INCREF(%s_wrapper)\n\n' % array)
    #raise NotImplementedError('multiple init are not supported')

    # Write the properties
    for key in variables:
        of.write(SPACING+'property %s:\n' % key)
        if key not in array_variables:
            of.write(2*SPACING+'def __get__(self):\n')
            of.write(3*SPACING+'return self._%s.%s\n' % (struct_name, key))
            of.write(2*SPACING+'def __set__(self, rhs):\n')
            of.write(3*SPACING+'self._%s.%s = rhs\n' % (struct_name, key))
        else:
            of.write(2*SPACING+'def __get__(self):\n')
            of.write(3*SPACING+'return self.%s_arr\n' % key)
            of.write(2*SPACING+'def __set__(self, rhs):\n')
            of.write(3*SPACING+'self.%s_arr[:] = rhs\n' % key)
        of.write('\n')

    # Add blank lines
    of.write('\n\n')


def write_imports(output_file):
    """TODO"""
    a = '''# Author: Gael Varoquaux
# License: BSD
from libc.stdlib cimport free
from cpython cimport PyObject, Py_INCREF

# Import the Python-level symbols of numpy
import numpy as np

# Import the C-level symbols of numpy
cimport numpy as np

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

cdef class ArrayWrapper:
    cdef void* data_ptr
    cdef int size
    cdef int type

    cdef set_data(self, int size, char* type, void* data_ptr):
        """ Set the data of the array

        This cannot be done in the constructor as it must recieve C-level
        arguments.

        Parameters:
        -----------
        size: int
            Length of the array.
        data_ptr: void*
            Pointer to the data

        """
        self.data_ptr = data_ptr
        self.size = size
        if type.find('int') != -1:
            self.type = np.NPY_INT
        elif type.find('float') != -1:
            self.type = np.NPY_FLOAT
        elif type.find('double') != -1:
            self.type = np.NPY_DOUBLE
        elif type.find('long') != -1:
            self.type = np.NPY_LONG

    def __array__(self):
        """ Here we use the __array__ method, that is called when numpy
        tries to get an array from the object."""
        cdef np.npy_intp shape[1]
        shape[0] = <np.npy_intp> self.size
        # Create a 1D array, of length 'size'
        ndarray = np.PyArray_SimpleNewFromData(1, shape,
            self.type, self.data_ptr)
        return ndarray

    def __dealloc__(self):
        """ Frees the array. This is called by Python when all the
        references to the object are gone. """
        free(<void*>self.data_ptr)\n\n'''
    output_file.write(a)


def create_logger():
    """Nothing"""

    logger = logging.getLogger('simple_example')
    #logger.setLevel(logging.DEBUG)
    logger.setLevel(logging.INFO)

    # create console handler and set level to debug
    console_handler = logging.StreamHandler()
    #console_handler.setLevel(logging.DEBUG)
    console_handler.setLevel(logging.INFO)

    # create formatter
    #formatter = logging.Formatter(
        #"%(asctime)s %(module)s: L%(lineno) 4s %(funcName) 15s"
        #" | %(levelname) -10s  --> %(message)s")
    formatter = colorlog.ColoredFormatter(
        "%(asctime)s %(module)s: L%(lineno) 4s %(blue)s%(funcName) 15s%(reset)s"
        " | %(log_color)s%(levelname) -10s  --> %(message)s%(reset)s",
        datefmt=None,
        reset=True,
        log_colors={
            'DEBUG':    'cyan',
            'INFO':     'green',
            'WARNING':  'yellow',
            'ERROR':    'red',
            'CRITICAL': 'red',
        })

    # add formatter to console_handler
    console_handler.setFormatter(formatter)

    # add console_handler to logger
    logger.addHandler(console_handler)

    return logger

if __name__ == "__main__":
    sys.exit(main())



"""
.. module:: test_class
    :synopsis: python script for testing CLASS using nose
.. moduleauthor:: Benjamin Audren <benjamin.audren@gmail.com>
.. credits:: Benjamin Audren, Thomas Tram
.. version:: 1.0

This is a python script for testing CLASS and its wrapper Classy using nose.
To run the test suite, type
nosetests test_class.py
If you want to extract the problematic input parameters at a later stage,
you should type
nosetests test_class.py 1>stdoutfile 2>stderrfile
and then use the python script extract_errors.py on the stderrfile.

When adding a new input parameter to CLASS (by modifying input.c), you
should also include tests of this new input. You will be in one of the
two cases:
1:  The new input is supposed to be compatible with any existing input.
    This is the standard case when adding a new species for instance.
2:  The new input is incompatible with one of the existing inputs. This
    would be the case if you have added (or just want to test) some other
    value of an already defined parameter. (Maybe you have allowed for
    negative mass neutrinos and you want to test CLASS using a negative mass.)

In case 1, you must add an entry in the CLASS_INPUT dictionary:
CLASS_INPUT['Mnu'] = (
    [{'N_eff': 0.0, 'N_ncdm': 1, 'm_ncdm': 0.06, 'deg_ncdm': 3.0},
     {'N_eff': 1.5, 'N_ncdm': 1, 'm_ncdm': 0.03, 'deg_ncdm': 1.5}],
    'normal')
The key 'Mnu' is not being used in the code, so its purpose is just to
describe the entry to the reader.
the value is a 2-tuple where the first entry [{},{},...,{}] is an array of
dictionaries containg the actual input to CLASS. The second entry is a keyword
which can be either 'normal' or 'power'. It tells the script how this input
will be combined with other inputs.

What does 'normal' and 'power' mean?
If an entry has the 'power' keyword, it will be combined with any other entry.
If an entry has the 'normal' keyword, it will not be combined with any other
entry having the 'normal' keyword, but it will be combined with all entries
carrying the 'power keyword.
Beware that the number of tests grow a lot when using the 'power' keyword.

In case 2, you should find the relevant entry and just add a new dictionary
to the array. E.g. if you want to test some negative mass model you should add
{'N_ncdm': 1, 'm_ncdm': -0.1, 'deg_ncdm': 1.0}

How are default parameters handled?
Any input array implicitly contains the empty dictionary. That means that if
Omega_k:0.0 is the default value, writing
CLASS_INPUT['Curvature'] = (
    [{'Omega_k': 0.01},
     {'Omega_k': -0.01}],
    'normal')
will test the default value Omega_k=0.0 along with the two specified models.

How to deal with inconsistent input?
Sometimes a specific feature requires the presence of another input parameter.
For instance, if we ask for tensor modes we must have temperature and/or
polarisation in the output. If not, CLASS is supposed to fail during the
evaluation of the input module and return an error message. This fail is the
correct behaviour of CLASS. To implement such a case, modify the function
has_incompatible_input(self)

Comparing output: When the flag 'COMPARE_OUTPUT_GAUGE' is set to true, the code will
rerun CLASS for each case under Newtonian gauge and then compare Cl's and
matter power spectrum. If the two are not close enough, it will generate a
PDF plot of this and save it in the 'fail' folder.
"""
from __future__ import print_function
from six import iteritems
#from future.utils import iteritems
import matplotlib as mpl
mpl.use('Agg')

import itertools
import matplotlib.pyplot as plt
import numpy as np
import os
import shutil
import unittest

from classy import Class
from classy import CosmoSevereError
from math import log10
from matplotlib.offsetbox import AnchoredText
from nose.plugins.attrib import attr
from parameterized import parameterized

# Customise test by reading environment variables
CLASS_VERBOSE = bool(int(os.getenv('CLASS_VERBOSE', '0'))) # Print output from CLASS?
COMPARE_OUTPUT_GAUGE = bool(int(os.getenv('COMPARE_OUTPUT_GAUGE', '0'))) # Compare synchronous and Newtonian gauge outputs?
COMPARE_OUTPUT_REF = bool(int(os.getenv('COMPARE_OUTPUT_REF', '0'))) # Compare classy with classyref?
POWER_ALL = bool(int(os.getenv('POWER_ALL', '0'))) # Combine every extension with each other? (Very slow!)
VALGRIND_MODE = bool(int(os.getenv('VALGRIND_MODE', '0')))

TEST_LEVEL = int(os.getenv('TEST_LEVEL', '0')) # 0 <= TEST_LEVEL <= 3

if COMPARE_OUTPUT_REF:
    try:
        import classyref
    except:
        COMPARE_OUTPUT_REF = False
# Define bounds on the relative and absolute errors of C(l) and P(k)
# between reference, Newtonian and Synchronous gauge
COMPARE_CL_RELATIVE_ERROR = 3e-3
COMPARE_CL_RELATIVE_ERROR_GAUGE = 5*3e-3
COMPARE_CL_ABSOLUTE_ERROR = 1e-20
COMPARE_PK_RELATIVE_ERROR = 1e-2
COMPARE_PK_RELATIVE_ERROR_GAUGE = 5*1e-2
COMPARE_PK_ABSOLUTE_ERROR = 1e-20
COMPARE_SHOOTING_RELATIVE_ERROR = 1e-5

# Dictionary of models to test the wrapper against. Each of these scenario will
# be run against all the possible output choices (nothing, tCl, mPk, etc...),
# with or without non-linearities.
# Never input the default value, as this will be **automatically** tested
# against. Indeed, when not specifying a field, CLASS takes the default input.
CLASS_INPUT = {}

# power = compare ALL settings with each of the entries
# normal = compare only the power settings with this entry
# onlyfull = compare only for 'full' output settings (see below)

CLASS_INPUT['Output_spectra'] = (
    [{'output': 'mPk', 'P_k_max_1/Mpc': 2},
     {'output': 'tCl'},
     {'output': 'tCl pCl lCl'},
     {'output': 'mPk tCl lCl', 'P_k_max_1/Mpc': 2},
     {'output': 'nCl sCl'},
     {'output': 'tCl pCl nCl sCl'}],
    'power')

CLASS_INPUT['Nonlinear'] = (
    [{'non_linear': 'halofit'},
    {'non_linear': 'hmcode'}],
    'power')

CLASS_INPUT['Lensing'] = (
    [{'lensing': 'yes'}],
    'power')

if TEST_LEVEL > 0:
    CLASS_INPUT['Mnu'] = (
        [{'N_ur': 0.0, 'N_ncdm': 1, 'm_ncdm': 0.06, 'deg_ncdm': 3.0},
        {'N_ur': 1.5, 'N_ncdm': 1, 'm_ncdm': 0.03, 'deg_ncdm': 1.5}],
        'normal')

if TEST_LEVEL > 1:
    CLASS_INPUT['Curvature'] = (
        [{'Omega_k': 0.01},
        {'Omega_k': -0.01}],
        'normal')

    CLASS_INPUT['modes'] = (
        [{'modes': 't'},
        {'modes': 's, t'}],
        'normal')

    CLASS_INPUT['Tensor_method'] = (
        [{'tensor method': 'exact'},
        {'tensor method': 'photons'}],
        'onlyfull')

if TEST_LEVEL > 2:
    CLASS_INPUT['Isocurvature_modes'] = (
        [{'ic': 'ad,nid,cdi', 'c_ad_cdi': -0.5}],
        'onlyfull')

    CLASS_INPUT['Scalar_field'] = (
        [{'Omega_scf': 0.1, 'attractor_ic_scf': 'yes',
        'scf_parameters': '10, 0, 0, 0'}],
        'onlyfull')

    CLASS_INPUT['Inflation'] = (
        [{'Pk_ini_type': 'inflation_V'},
        {'Pk_ini_type': 'inflation_H'}],
        'onlyfull')

    CLASS_INPUT['Shooting'] = (
        [{'sigma8': 0.8},
        {'S8': 0.8},
        {'theta_s_100': 1.04},
        {'Neff': 4.0},
        {'Neff': 2.0},
        {'Neff': 2.0,'N_ncdm':1,'m_ncdm':0.06},
        {'Neff': 3.0,'N_ncdm':1,'m_ncdm':0.06},
        {'Neff': 4.0,'N_ncdm':1,'m_ncdm':0.06},
        {'Neff': 4.0,'N_ncdm':1,'m_ncdm':0.06, 'theta_s_100':1.04,'sigma8':0.8},
        {'Neff': 4.0,'N_ncdm':1,'m_ncdm':0.06, 'theta_s_100':1.04,'S8':0.8},
        {'tau_reio':0.05}
        ],
        'onlyfull')
# TODO :: possibly another one with nonlinear=hmcode in the future, but for now this is more stable
FULL_SETTINGS = {'output':'mPk mTk vTk tCl pCl lCl sCl nCl','P_k_max_1/Mpc': 2,'lensing':'yes','modes':'s,t'}

if POWER_ALL:
    for k, v in iteritems(CLASS_INPUT):
        models, state = v
        CLASS_INPUT[k] = (models, 'power')
if VALGRIND_MODE:
    for k, v in iteritems(CLASS_INPUT):
        if k == 'Output_spectra':
            continue
        models, state = v
        CLASS_INPUT[k] = (models, 'normal')

INPUTPOWER = []
INPUTNORMAL = [{}]
INPUTONLYFULL = [{}]
for key, value in list(CLASS_INPUT.items()):
    models, state = value
    if state == 'power':
        INPUTPOWER.append([{}]+models)
    elif state == 'onlyfull':
        INPUTONLYFULL.extend(models)
    else:
        INPUTNORMAL.extend(models)

    PRODPOWER = list(itertools.product(*INPUTPOWER))

    DICTARRAY = []
    for normelem in INPUTNORMAL:
        for powelem in PRODPOWER:  # itertools.product(*modpower):
            temp_dict = normelem.copy()
            for elem in powelem:
                temp_dict.update(elem)
            DICTARRAY.append(temp_dict)

    for onlyfull in INPUTONLYFULL:
        temp_dict = onlyfull.copy()
        temp_dict.update(FULL_SETTINGS)
        DICTARRAY.append(temp_dict)

TUPLE_ARRAY = []
for e in DICTARRAY:
    TUPLE_ARRAY.append((e, ))


def powerset(iterable):
    xs = list(iterable)
    # note we return an iterator rather than a list
    return itertools.chain.from_iterable(
        itertools.combinations(xs, n) for n in range(1, len(xs)+1))

def custom_name_func(testcase_func, param_num, param):
    special_keys = ['N_ncdm','Pk_ini_type','ic','Neff','sigma8','S8','theta_s_100','tau_reio']
    somekeys = []
    for key in param.args[0]:
        if key in special_keys:
            somekeys.append(key)
        elif 'mega' in key:
            somekeys.append(key)

    res = '{}_{:04d}_{}'.format(
        testcase_func.__name__,
        param_num,
        parameterized.to_safe_name('_'.join(somekeys))
    )
    return res.strip('_')

class TestClass(unittest.TestCase):
    """
    Testing Class and its wrapper classy on different cosmologies

    To run it, do
    ~] nosetest test_class.py

    It will run many times Class, on different cosmological scenarios, and
    everytime testing for different output possibilities (none asked, only mPk,
    etc..)

    """
    @classmethod
    def setUpClass(cls):
        cls.faulty_figs_path = os.path.join(
            os.path.sep.join(os.path.realpath(__file__).split(
                os.path.sep)[:-1]),
            'faulty_figs')

        if os.path.isdir(cls.faulty_figs_path):
            shutil.rmtree(cls.faulty_figs_path)

        os.mkdir(cls.faulty_figs_path)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        """
        set up data used in the tests.
        setUp is called before each test function execution.
        """
        self.cosmo = Class()
        self.cosmo_newt = Class()

        if CLASS_VERBOSE:
            self.verbose = {
                'input_verbose': 1,
                'background_verbose': 1,
                'thermodynamics_verbose': 1,
                'perturbations_verbose': 1,
                'transfer_verbose': 1,
                'primordial_verbose': 1,
                'harmonic_verbose': 1,
                'fourier_verbose': 1,
                'lensing_verbose': 1,
                'distortions_verbose': 1,
                'output_verbose': 1,
            }
        else:
            self.verbose = {}
        self.scenario = {}

    def tearDown(self):
        self.cosmo.struct_cleanup()
        self.cosmo.empty()
        self.cosmo = 0
        self.cosmo_newt.struct_cleanup()
        self.cosmo_newt.empty()
        self.cosmo_newt = 0
        del self.scenario

    def poormansname(self, somedict):
        string = "_".join(
            [k+'='+str(v)
             for k, v in list(somedict.items())])
        string = string.replace('/', '%')
        string = string.replace(',', '')
        string = string.replace(' ', '')
        return string

    @parameterized.expand(TUPLE_ARRAY, doc_func=custom_name_func, custom_name_func=custom_name_func)
    @attr('dump_ini_files')
    def test_Valgrind(self, inputdict):
        """Dump files"""
        self.scenario.update(inputdict)
        self.name = self._testMethodName
        if self.has_incompatible_input():
            return
        path = os.path.join(self.faulty_figs_path, self.name)
        self.store_ini_file(path)
        self.scenario.update({'gauge':'Newtonian'})
        self.store_ini_file(path + 'N')

    @parameterized.expand(TUPLE_ARRAY, doc_func=custom_name_func, custom_name_func=custom_name_func)
    @attr('test_scenario')
    def test_scenario(self, inputdict):
        """Test scenario"""
        self.scenario.update(inputdict)
        self.name = self._testMethodName
        self.cosmo.set(dict(itertools.chain(self.verbose.items(), self.scenario.items())))

        cl_dict = {
            'tCl': ['tt'],
            'lCl': ['pp'],
            'pCl': ['ee', 'bb'],
            'nCl': ['dd'],
            'sCl': ['ll'],
        }

        # 'lensing' is always set to yes. Therefore, trying to compute 'tCl' or
        # 'pCl' will fail except if we also ask for 'lCl'.
        if self.has_incompatible_input():
            self.assertRaises(CosmoSevereError, self.cosmo.compute)
            return
        else:
            self.cosmo.compute()

        self.assertTrue(
            self.cosmo.state,
            "Class failed to go through all __init__ methods")
        # Depending
        if 'output' in self.scenario:
            # Positive tests of raw cls
            output = self.scenario['output']
            for elem in output.split():
                if elem in cl_dict:
                    for cl_type in cl_dict[elem]:
                        is_density_cl = (elem == 'nCl' or elem == 'sCl')
                        if is_density_cl:
                            cl = self.cosmo.density_cl(100)
                        else:
                            cl = self.cosmo.raw_cl(100)
                        if is_density_cl:
                          self.assertIsNotNone(cl, "density_cl returned nothing")
                          if cl_type=='ell':
                            cl_length = np.shape(cl[cl_type])[0]
                            self.assertEqual(cl_length, 101, "density_cl returned wrong ell size")
                            continue
                          for val in cl[cl_type]:
                            cl_length = np.shape(cl[cl_type][val])[0]
                            self.assertEqual(cl_length, 101, "density_cl returned wrong size")
                        else:
                          cl_length = np.shape(cl[cl_type])[0]
                          self.assertIsNotNone(cl, "raw_cl returned nothing")
                          self.assertEqual(cl_length, 101, "raw_cl returned wrong size")
                if elem == 'mPk':
                    pk = self.cosmo.pk(0.1, 0)
                    self.assertIsNotNone(pk, "pk returned nothing")
            # Negative tests of output functions
            if not any([elem in cl_dict for elem in output.split()]):
                # testing absence of any Cl
                self.assertRaises(CosmoSevereError, self.cosmo.raw_cl, 100)
            if 'mPk' not in output.split():
                # testing absence of mPk
                self.assertRaises(CosmoSevereError, self.cosmo.pk, 0.1, 0)

        if COMPARE_OUTPUT_REF or COMPARE_OUTPUT_GAUGE:
            # Now compute same scenario in Newtonian gauge
            self.cosmo_newt.set(
                dict(list(self.verbose.items())+list(self.scenario.items())))
            self.cosmo_newt.set({'gauge': 'newtonian'})
            self.cosmo_newt.compute()

        if COMPARE_OUTPUT_GAUGE:
            # Compare synchronous and Newtonian gauge
            self.assertTrue(
                self.cosmo_newt.state,
                "Class failed to go through all __init__ methods in Newtonian gauge")

            self.compare_output(self.cosmo, "Synchronous", self.cosmo_newt, 'Newtonian', COMPARE_CL_RELATIVE_ERROR_GAUGE, COMPARE_PK_RELATIVE_ERROR_GAUGE)

        if COMPARE_OUTPUT_REF:
            # Compute reference models in both gauges and compare
            cosmo_ref = classyref.Class()
            cosmo_ref.set(self.cosmo.pars)
            cosmo_ref.compute()
            status = self.compare_output(cosmo_ref, "Reference", self.cosmo, 'Synchronous', COMPARE_CL_RELATIVE_ERROR, COMPARE_PK_RELATIVE_ERROR)
            assert status, 'Reference comparison failed in Synchronous gauge!'

            cosmo_ref = classyref.Class()
            cosmo_ref.set(self.cosmo_newt.pars)
            cosmo_ref.compute()
            self.compare_output(cosmo_ref, "Reference", self.cosmo_newt, 'Newtonian', COMPARE_CL_RELATIVE_ERROR, COMPARE_PK_RELATIVE_ERROR)
            assert status, 'Reference comparison failed in Newtonian gauge!'

    def has_incompatible_input(self):

        should_fail = False

        # If we have tensor modes, we must have one tensor observable,
        # either tCl or pCl.
        if has_tensor(self.scenario):
            if 'output' not in self.scenario:
                should_fail = True
            else:
                output = self.scenario['output'].split()
                if 'tCl' not in output and 'pCl' not in output:
                    should_fail = True

        # If we have specified lensing, we must have lCl in output,
        # otherwise lensing will not be read (which is an error).
        if 'lensing' in self.scenario:
            if 'output' not in self.scenario:
                should_fail = True
            else:
                output = self.scenario['output'].split()
                if 'lCl' not in output:
                    should_fail = True
                elif 'tCl' not in output and 'pCl' not in output:
                    should_fail = True

        # If we have specified a tensor method, we must have tensors.
        if 'tensor method' in self.scenario:
            if not has_tensor(self.scenario):
                should_fail = True

        # If we have specified non_linear, we must have some form of
        # perturbations output.
        if 'non_linear' in self.scenario:
            if 'output' not in self.scenario:
                should_fail = True

        # If we ask for Cl's of lensing potential, number counts or cosmic shear, we must have scalar modes.
        # The same applies to density and velocity transfer functions and the matter power spectrum:
        if 'output' in self.scenario and 'modes' in self.scenario and self.scenario['modes'].find('s') == -1:
            requested_output_types = set(self.scenario['output'].split())
            for scalar_output_type in ['lCl', 'nCl', 'dCl', 'sCl', 'mPk', 'dTk', 'mTk', 'vTk']:
                if scalar_output_type in requested_output_types:
                    should_fail = True
                    break

        # If we specify initial conditions (for scalar modes), we must have
        # perturbations and scalar modes.
        if 'ic' in self.scenario:
            if 'modes' in self.scenario and self.scenario['modes'].find('s') == -1:
                should_fail = True
            if 'output' not in self.scenario:
                should_fail = True

        # If we use inflation module, we must have scalar modes,
        # tensor modes, no vector modes and we should only have adiabatic IC:
        if 'Pk_ini_type' in self.scenario and self.scenario['Pk_ini_type'].find('inflation') != -1:
            if 'modes' not in self.scenario:
                should_fail = True
            else:
                if self.scenario['modes'].find('s') == -1:
                    should_fail = True
                if self.scenario['modes'].find('v') != -1:
                    should_fail = True
                if self.scenario['modes'].find('t') == -1:
                    should_fail = True
            if 'ic' in self.scenario and self.scenario['ic'].find('i') != -1:
                should_fail = True
            if 'non_linear' in self.scenario and self.scenario['non_linear'].find('hmcode') != -1:
                should_fail = True


        return should_fail

    def compare_output(self, reference, reference_name, candidate, candidate_name, rtol_cl, rtol_pk):
        status_pass = True
        for elem in ['raw_cl', 'lensed_cl']:
            # Try to get the elem, but if they were not computed, a
            # CosmoComputeError should be raised. In this case, ignore the
            # whole block.
            try:
                to_test = getattr(candidate, elem)()
            except CosmoSevereError:
                continue
            ref = getattr(reference, elem)()
            for key, value in list(ref.items()):
                if key != 'ell':
                    # For all self spectra, try to compare allclose
                    if key[0] == key[1]:
                        # If it is a 'dd' or 'll', it is a dictionary.
                        if isinstance(value, dict):
                            for subkey in value:
                                try:
                                    np.testing.assert_allclose(
                                        value[subkey],
                                        to_test[key][subkey],
                                        rtol=rtol_cl,
                                        atol=COMPARE_CL_ABSOLUTE_ERROR)
                                except AssertionError:
                                    self.cl_faulty_plot(elem + "_" + key, value[subkey][2:], reference_name, to_test[key][subkey][2:], candidate_name, rtol_cl)
                                except TypeError:
                                    self.cl_faulty_plot(elem + "_" + key, value[subkey][2:], reference_name, to_test[key][subkey][2:], candidate_name, rtol_cl)
                        else:
                            try:
                                np.testing.assert_allclose(
                                    value,
                                    to_test[key],
                                    rtol=rtol_cl,
                                    atol=COMPARE_CL_ABSOLUTE_ERROR)
                            except (AssertionError, TypeError) as e:
                                self.cl_faulty_plot(elem + "_" + key, value[2:], reference_name, to_test[key][2:], candidate_name, rtol_cl)
                                status_pass = False
                    # For cross-spectra, as there can be zero-crossing, we
                    # instead compare the difference.
                    else:
                        # First, we multiply each array by the biggest value
                        norm = max(
                            np.abs(value).max(), np.abs(to_test[key]).max())
                        value *= norm
                        to_test[key] *= norm
                        try:
                            np.testing.assert_array_almost_equal(
                                value, to_test[key], decimal=3)
                        except AssertionError:
                            self.cl_faulty_plot(elem + "_" + key, value[2:], reference_name, to_test[key][2:], candidate_name, rtol_cl)
                            status_pass = False

        if 'output' in self.scenario:
            if self.scenario['output'].find('mPk') != -1:
                # testing equality of Pk
                k = np.logspace(-2, log10(self.scenario['P_k_max_1/Mpc']), 50)
                reference_pk = np.array([reference.pk(elem, 0) for elem in k])
                candidate_pk = np.array([candidate.pk(elem, 0) for elem in k])
                try:
                    np.testing.assert_allclose(
                        reference_pk,
                        candidate_pk,
                        rtol=rtol_pk,
                        atol=COMPARE_PK_ABSOLUTE_ERROR)
                except AssertionError:
                    self.pk_faulty_plot(k, reference_pk, reference_name, candidate_pk, candidate_name, rtol_pk)
                    status_pass = False

        if 'output' in self.scenario and self.scenario['output'].find('mPk') != -1:
            ref_sigma8 = reference.sigma8()
            can_sigma8 = candidate.sigma8()
            ref_sigma = reference.sigma(8, 0, h_units=True)
            can_sigma = candidate.sigma(8, 0, h_units=True)
            try:
                np.testing.assert_allclose([ref_sigma8, ref_sigma, can_sigma8],
                                           [can_sigma8, can_sigma, can_sigma],
                                           rtol=rtol_pk,
                                           atol=COMPARE_PK_ABSOLUTE_ERROR)
            except AssertionError:
                status_pass = False

        if 'sigma8' in self.scenario:
            if not np.isclose(self.scenario['sigma8'], candidate.sigma8(), rtol=COMPARE_SHOOTING_RELATIVE_ERROR, atol=0):
              status_pass = False
            if not np.isclose(candidate.sigma8(), reference.sigma8(), rtol=COMPARE_SHOOTING_RELATIVE_ERROR, atol=0):
              status_pass = False

        if 'Neff' in self.scenario:
            if not np.isclose(self.scenario['Neff'], candidate.Neff(), rtol=COMPARE_SHOOTING_RELATIVE_ERROR, atol=0):
              status_pass = False
            if not np.isclose(candidate.Neff(), reference.Neff(), rtol=COMPARE_SHOOTING_RELATIVE_ERROR, atol=0):
              status_pass = False

        if 'theta_s_100' in self.scenario:
            if not np.isclose(self.scenario['theta_s_100'], candidate.theta_s_100(), rtol=COMPARE_SHOOTING_RELATIVE_ERROR, atol=0):
              status_pass = False
            if not np.isclose(candidate.theta_s_100(), reference.theta_s_100(), rtol=COMPARE_SHOOTING_RELATIVE_ERROR, atol=0):
              status_pass = False
        return status_pass

    def store_ini_file(self, path):
        parameters = dict(list(self.verbose.items()) + list(self.scenario.items()))
        with open(path + '.ini', 'w') as param_file:
            param_file.write('# ' + str(parameters) + '\n')
            if len(parameters) == 0:
                # CLASS complains if the .ini file does not do anything.
                param_file.write('write warnings = yes\n')
            for key, value in list(parameters.items()):
                param_file.write(key + " = " + str(value)+ '\n')

    def cl_faulty_plot(self, cl_type, reference, reference_name, candidate, candidate_name, rtol):
        path = os.path.join(self.faulty_figs_path, self.name)
        fig, axes = plt.subplots(2, 1, sharex=True)
        ell = np.arange(max(np.shape(candidate))) + 2
        factor = ell*(ell + 1)/(2*np.pi) if cl_type[-2:] != 'pp' else ell**5
        axes[0].plot(ell, factor*reference, label=reference_name)
        axes[0].plot(ell, factor*candidate, label=candidate_name)
        axes[1].semilogy(ell, 100*abs(candidate/reference - 1), label=cl_type)
        axes[1].axhline(y=100*rtol, color='k', ls='--')

        axes[-1].set_xlabel(r'$\ell$')
        if cl_type[-2:] == 'pp':
            axes[0].set_ylabel(r'$\ell^5 C_\ell^\mathrm{{{_cl_type}}}$'.format(_cl_type=cl_type[-2:].upper()))
        else:
            axes[0].set_ylabel(r'$\ell(\ell + 1)/(2\pi)C_\ell^\mathrm{{{_cl_type}}}$'.format(_cl_type=cl_type[-2:].upper()))
        axes[1].set_ylabel('Relative error [%]')

        for ax in axes:
            ax.legend(loc='upper right')

        fig.tight_layout()
        fname = '{}_{}_{}_vs_{}.pdf'.format(path, cl_type, reference_name, candidate_name)
        fig.savefig(fname, bbox_inches='tight')
        plt.close(fig)

        # Store parameters (contained in self.scenario) to text file
        self.store_ini_file(path)

    def pk_faulty_plot(self, k, reference, reference_name, candidate, candidate_name, rtol):
        path = os.path.join(self.faulty_figs_path, self.name)

        fig, axes = plt.subplots(2, 1, sharex=True)
        axes[0].loglog(k, k**1.5*reference, label=reference_name)
        axes[0].loglog(k, k**1.5*candidate, label=candidate_name)
        axes[0].legend(loc='upper right')

        axes[1].loglog(k, 100*np.abs(candidate/reference - 1))
        axes[1].axhline(y=100*rtol, color='k', ls='--')

        axes[-1].set_xlabel(r'$k\quad [\mathrm{Mpc}^{-1}]$')
        axes[0].set_ylabel(r'$k^\frac{3}{2}P(k)$')
        axes[1].set_ylabel(r'Relative error [%]')

        fig.tight_layout()
        fname = path + '_pk_{}_vs_{}.pdf'.format(reference_name, candidate_name)
        fig.savefig(fname, bbox_inches='tight')
        plt.close(fig)

        # Store parameters (contained in self.scenario) to text file
        self.store_ini_file(path)

def has_tensor(input_dict):
    if 'modes' in input_dict:
        if input_dict['modes'].find('t') != -1:
            return True
    else:
        return False
    return False

if __name__ == '__main__':
    toto = TestClass()
    unittest.main()



# From the dumped stdout and stderr of a nosetests test_class.py, extract all
# the failed steps.
# Usage: python extract_errors.py output
from __future__ import print_function
import sys
import os


def main(path):
    """
    Create a shorter file containing only the errors from nosetests

    """
    assert os.path.isfile(path) is True

    trimmed_path = path + '_errors'
    destination = open(trimmed_path, 'w')
    contains_error = False
    with open(path, 'r') as source:
        text = source.readlines()
        start = 0
        for index, line in enumerate(text):
            if line.find('------------------') != -1:
                if text[index+2].find('----------------') != -1:
                    stop = index-1
                    # Check that an error is contained
                    if stop > 0:
                        for i in range(start, stop+1):
                            if text[i].startswith('E'):
                                contains_error = True
                    if contains_error:
                        print('Found an error')
                        for i in range(start, stop+1):
                            print(text[i], end=' ')
                            destination.write(text[i])
                    start = index
                    contains_error = False
                elif text[index+2].find('=================') != -1:
                    break
                else:
                    pass
    destination.close()


if __name__ == "__main__":
    print(sys.argv)
    if len(sys.argv) != 2:
        print('Please specify the output file to analyse')
        exit()
    else:
        main(sys.argv[-1])



from setuptools import setup, Extension
from Cython.Distutils import build_ext

import numpy as np
import os
import subprocess as sbp
import sys
from pip._internal.commands.install import decide_user_install

def binaries_directory():
    import site
    """Return the installation directory, or None"""
    user_install = decide_user_install(use_user_site = True if ('--user' in sys.argv) else None, prefix_path = any(['--prefix' in x for x in sys.argv]), target_dir=any(['--target' in x for x in sys.argv]), root_path = None, isolated_mode = True if ('--no-build-isolation' in sys.argv) else None)

    if user_install:
        paths = (site.getusersitepackages(),)
    else:
        py_version = '%s.%s' % (sys.version_info[0], sys.version_info[1])
        paths = (s % (py_version) for s in (
            sys.prefix + '/lib/python%s/dist-packages',
            sys.prefix + '/lib/python%s/site-packages',
            sys.prefix + '/local/lib/python%s/dist-packages',
            sys.prefix + '/local/lib/python%s/site-packages',
            '/Library/Python/%s/site-packages',
        ))

    for path in paths:
        if os.path.exists(path):
            return path
    print('no installation path found', file=sys.stderr)
    return None


# Get the root folder of the CLASS installation -- this setup.py should be in that folder
root_folder = os.path.dirname(os.path.abspath(__file__))

# Recover the gcc compiler
GCCPATH_STRING = sbp.Popen(
    ['gcc', '-print-libgcc-file-name'],
    stdout=sbp.PIPE).communicate()[0]
GCCPATH = os.path.normpath(os.path.dirname(GCCPATH_STRING)).decode()

liblist = ["class"]
MVEC_STRING = sbp.Popen(
    ['gcc', '-lmvec'],
    stderr=sbp.PIPE).communicate()[1]
if b"mvec" not in MVEC_STRING:
    liblist += ["mvec","m"]

# define absolute paths
include_folder = os.path.join(root_folder, "include")
classy_folder = os.path.join(root_folder, "python")
heat_folder = os.path.join(os.path.join(root_folder, "external"),"heating")
recfast_folder = os.path.join(os.path.join(root_folder, "external"),"RecfastCLASS")
hyrec_folder = os.path.join(os.path.join(root_folder, "external"),"HyRec2020")

# Recover the CLASS version
with open(os.path.join(include_folder, 'common.h'), 'r') as v_file:
    for line in v_file:
        if line.find("_VERSION_") != -1:
            # get rid of the " and the v
            VERSION = line.split()[-1][2:-1]+".0" # Always set a sub-version number here for subsequent uploads (!)
            break

# Define cython extension and fix Python version
classy_ext = Extension("classy", [os.path.join("python", "classy.pyx")],
                       include_dirs=[np.get_include(), include_folder, heat_folder, recfast_folder, hyrec_folder],
                       libraries=liblist,
                       library_dirs=[root_folder, GCCPATH],
                       language="c++",
                       extra_compile_args=["-std=c++11"],
                       depends=["libclass.a","python/cclassy.pxd"]
                       )

classy_ext.cython_directives = {'language_level': "3" if sys.version_info.major>=3 else "2"}


# We need to put the actual files somewhere (e.g. the BBN file -- the easiest is just to copy the full class folder (although in the future we could imagine just copying the necessary files)

# 1. Check where to install class to
path_install = os.path.join(binaries_directory(),"class")
print("Selected installation path : ", path_install)


# 2. Check what files to include
def package_files(directory):
    paths = []
    direcs = []
    wanted_paths = {os.path.join(directory, d) for d in ["tools", "source", "main", "python", "include"]}
    for (path, directories, filenames) in os.walk(directory):
        # Only include those directories that we actually want
        if (path in wanted_paths or
            (path.startswith(os.path.join(directory,"external")) and not 'RealSpaceInterface' in path)):
          print("INCLUDING", path, filenames)
          for filename in filenames:
              paths.append(os.path.join(path, filename))
    return paths

pck_files = package_files(".")
pck_files.append("./Makefile")
# Debug print (only occurs when executed with the -v option)
print("Included files : ", pck_files)


# Make a custom builder in order to compile the C code as well, using the makefile
class classy_builder(build_ext):

    def build_extension(self, ext):
      # Make sure to put the current python version into the 'PYTHON' variable
      env = os.environ.copy()
      if not 'PYTHON' in env:
        env['PYTHON'] = sys.executable

      env['CLASSDIR'] = path_install

      # Compile the C code only
      returncode = sbp.call(["make","libclass.a","-j"], env=env)
      if returncode!=0:
        raise RuntimeError("Unknown error occurred -- the Makefile compilation of class failed (return code %i). Run the installation with '-v' and check why this is the case from the makefile command output"%returncode)
      super().build_extension(ext)

# Finally, perform the actual setup
setup(
    name='classy',
    version=VERSION,
    description='Python interface to the Cosmological Boltzmann code CLASS',
    url='http://www.class-code.net',
    cmdclass={'build_ext': classy_builder},
    ext_modules=[classy_ext],
    packages = ["class"],
    package_dir={"class":""},
    package_data={'class': pck_files},
    include_package_data=True
)



#!/usr/bin/env python

import numpy as np
import sys
import scipy.interpolate as sciint
from numpy.linalg import norm as vector_norm
from numpy.linalg import eigh as eigen_vals_vecs
import os
import matplotlib.pyplot as plt

# Read inputs
if(len(sys.argv)==14):
  sd_detector_name = sys.argv[1]
  sd_detector_nu_min = eval(sys.argv[2])
  sd_detector_nu_max = eval(sys.argv[3])
  sd_detector_nu_delta = eval(sys.argv[4])
  sd_detector_bin_number = eval(sys.argv[5])
  sd_z_min = eval(sys.argv[6])
  sd_z_max = eval(sys.argv[7])
  sd_z_size = eval(sys.argv[8])
  sd_detector_delta_Ic = eval(sys.argv[9])
  sd_PCA_size = eval(sys.argv[10])
  z_th = eval(sys.argv[11])
  DI_units = eval(sys.argv[12])   # = 2.70062634e-18
  x_to_nu = eval(sys.argv[13])    # = 56.7798
  has_noisefile = False
elif(len(sys.argv)==11):
  sd_detector_name = sys.argv[1]
  sd_external_path = sys.argv[2]
  sd_noisefile_name = sys.argv[3]
  sd_z_min = eval(sys.argv[4])
  sd_z_max = eval(sys.argv[5])
  sd_z_size = eval(sys.argv[6])
  sd_PCA_size = eval(sys.argv[7])
  z_th = eval(sys.argv[8])
  DI_units = eval(sys.argv[9])   # = 2.70062634e-18
  x_to_nu = eval(sys.argv[10])    # = 56.7798
  has_noisefile = True
else:
  raise Exception("generate_PCA_files.py received invalid input arguments")

def PCA_string_to_array(line,delimiter=" "):
  line = line.replace("\n","")
  if delimiter is not "\t":
    line = line.replace("\t","")
  return np.array([float(x) for x in line.split(delimiter) if (x is not "" and x is not " ")])

def read_noisefile(filename):
  with open(filename) as det_noise:
    header = True
    while(header):
      line = det_noise.readline()
      if(line.startswith("#")):
        continue
      header=False
    Nrows,Ncols = PCA_string_to_array(line)
    #Skip first line containing Nx,Ncols
    line = det_noise.readline()
    cols = []
    while(line):
      cols.append(PCA_string_to_array(line))
      line = det_noise.readline()
    cols = np.array(cols).T
    assert(int(Ncols)==len(cols))
    assert(int(Nrows)==len(cols[0]))
    return len(cols[0]),cols[0]/x_to_nu,cols[1]*1e-26

dir_path = os.path.dirname(os.path.realpath(__file__))

# Read external file Greens_data.dat
readfile = "Greens_data.dat"
with open(os.path.join(dir_path,readfile)) as f:
  # Read the header first
  header = True
  while(header):
    line = f.readline()
    if(line.startswith("#")):
      continue
    # The first line of the header without the "#" is still part of the header
    header=False

  # Read the first line specifying z
  Greens_z = PCA_string_to_array(f.readline())
  Greens_Nz = len(Greens_z)
  Greens_lnz = np.log(Greens_z+1.)

  # Read T_ini,T_last and rho
  Greens_T_ini = PCA_string_to_array(f.readline())
  Greens_T_last = PCA_string_to_array(f.readline())
  Greens_drho = PCA_string_to_array(f.readline())

  # Calculate the difference in Temperature
  Greens_dT = (Greens_T_last-Greens_T_ini)/Greens_T_ini

  # Read the rest of the file
  done = False
  Greens_data_full = []
  while(not done):
    line = f.readline()
    if(not line):
      done = True
    else:
      Greens_data_full.append(PCA_string_to_array(line))
  Greens_data_full = np.array(Greens_data_full).T

  # Seperate the rest of the data into x, Green(z,x) and the blackbody
  Greens_x = Greens_data_full[0]
  Greens_Nx = len(Greens_x)
  Greens_G_th = Greens_data_full[1:Greens_Nz+1]
  Greens_blackbody = Greens_data_full[Greens_Nz+1]

  # Spline Greens function for interpolation
  Greens_G_th_Spline = [None for index_x_old in range(Greens_Nx)]
  for index_x_old in range(Greens_Nx):
    Greens_G_th_Spline[index_x_old] = sciint.CubicSpline(Greens_lnz,Greens_G_th[:,index_x_old])

  # Spline Greens dT for interpolation
  Greens_T_ini_Spline = sciint.CubicSpline(Greens_lnz,Greens_T_ini)
  Greens_T_last_Spline = sciint.CubicSpline(Greens_lnz,Greens_T_last)
  Greens_dT_Spline = sciint.CubicSpline(Greens_lnz,Greens_dT)
  Greens_drho_Spline = sciint.CubicSpline(Greens_lnz,Greens_drho)

  # Define new z and x arrays
  Nz_arr = sd_z_size
  z_arr = np.logspace(np.log10(sd_z_min),np.log10(sd_z_max),Nz_arr)
  lnz_arr = np.log(z_arr+1.)

  if has_noisefile:
    Nx_arr,x_arr,deltaIc_arr = read_noisefile(os.path.join(sd_external_path,sd_noisefile_name))
  else:
    Nx_arr = sd_detector_bin_number+1
    x_arr = np.linspace(sd_detector_nu_min/x_to_nu,sd_detector_nu_max/x_to_nu,Nx_arr)

  # Define visibility function
  #bb_vis = np.exp(-(z_arr/2.021e6)**2.5)
  bb_vis = np.exp(-(z_arr/z_th)**2.5)

  # The Gth file of Chluba subtracts away some part of the G_T distortion into a shift from T_ini to T_last
  # Here we calculate backwards, and obtain the shift of f_g due to the internal dT
  df_g = Greens_dT_Spline(lnz_arr)/Greens_drho_Spline(lnz_arr)

  # Initialize spectral shapes
  G_th = np.zeros((Nx_arr,Nz_arr))
  DI_T_shift = np.zeros((Nx_arr,Nz_arr))
  Gdist = np.zeros(Nx_arr)
  Ydist = np.zeros(Nx_arr)
  Mdist = np.zeros(Nx_arr)

  # Interpolate Green's function
  index_x_old = 0
  for index_x_new,x in enumerate(x_arr):
    # Define spectral shapes
    Gdist[index_x_new] = (x**4*np.exp(x)/(np.exp(x)-1)**2)*DI_units*1.0e18
    Ydist[index_x_new] = Gdist[index_x_new]*(x/np.tanh(x/2.)-4.)
    Mdist[index_x_new] = Gdist[index_x_new]*(1./2.19229-1./x)

    x_s = Greens_T_ini_Spline(lnz_arr)/Greens_T_last_Spline(lnz_arr)*x
    x_z = x*lnz_arr/lnz_arr
    DI_T_shift[index_x_new,:] = DI_units*1.0e26*x_z**3.*(np.exp(-x_s)/(1.-np.exp(-x_s))-np.exp(-x_z)/(1.-np.exp(-x_z)))/Greens_drho_Spline(lnz_arr)

    try:
      # Find position in xarray
      while(x>Greens_x[index_x_old]):
        index_x_old += 1
      # Linear interpolation in x
      frac = (x-Greens_x[index_x_old])/(Greens_x[index_x_old+1]-Greens_x[index_x_old])

      # Cubic interpolation for all values of z
      lowx_vals = Greens_G_th_Spline[index_x_old](lnz_arr)
      highx_vals = Greens_G_th_Spline[index_x_old+1](lnz_arr)

      G_th[index_x_new,:] = (lowx_vals*(1.-frac)+highx_vals*frac)
      G_th[index_x_new,:] *= bb_vis*1.e-8
      G_th[index_x_new,:] += DI_T_shift[index_x_new,:]*1.e-8
      #G_th[index_x_new,:] += Gdist[index_x_new]*df_g
    except:
      raise ValueError("{} is not in the file range [{},{}] for file '{}'".format(x,Greens_x[0],Greens_x[-1],readfile))

  # Begin orthonormlization
  # Y distortion
  e_Y = Ydist/vector_norm(Ydist)
  M_Y = np.dot(e_Y,Mdist)
  G_Y = np.dot(e_Y,Gdist)
  # Mu distortion
  Mperp = Mdist-M_Y*e_Y
  e_M = Mperp/vector_norm(Mperp)
  G_M = np.dot(e_M,Gdist)
  # G distortion
  Gperp = Gdist-G_Y*e_Y-G_M*e_M
  e_G = Gperp/vector_norm(Gperp)

  f_g = np.zeros(Nz_arr)
  f_mu = np.zeros(Nz_arr)
  f_y = np.zeros(Nz_arr)
  # Now, factorize G into orthonormal subspace
  for index_z in range(Nz_arr):
    # Compute non-normalized components
    f_g[index_z]  = (np.dot(G_th[:,index_z],e_G))/vector_norm(Gperp)
    f_mu[index_z] = (np.dot(G_th[:,index_z],e_M)-G_M*f_g[index_z])/vector_norm(Mperp)
    f_y[index_z]  = (np.dot(G_th[:,index_z],e_Y)-M_Y*f_mu[index_z]-G_Y*f_g[index_z])/vector_norm(Ydist)

  # Now we can re-normalize our functions and add the shift
  J_g = 4.*f_g
  J_mu = f_mu/1.401
  J_y  = 4.*f_y

  # Calculate non-normalized residual
  Residual = np.zeros((Nx_arr,Nz_arr))
  for index_x in range(Nx_arr):
    for index_z in range(Nz_arr):
      Residual[index_x,index_z] = G_th[index_x,index_z]-Gdist[index_x]*f_g[index_z]-Ydist[index_x]*f_y[index_z]-Mdist[index_x]*f_mu[index_z]

  # Calculate Fisher matrix
  Fisher = np.zeros((Nz_arr,Nz_arr))
  delta_ln_z = np.log(z_arr[1])-np.log(z_arr[0])
  for index_za in range(Nz_arr):
    for index_zb in range(Nz_arr):
      if has_noisefile:
        Fisher[index_za,index_zb] = np.sum(Residual[:,index_za]*Residual[:,index_zb]*pow(delta_ln_z/deltaIc_arr[:]*1.e8,2.))
      else:
        Fisher[index_za,index_zb] = np.sum(Residual[:,index_za]*Residual[:,index_zb]*pow(delta_ln_z/sd_detector_delta_Ic*1.e8,2.))

  # Solve eigenvalue problem
  eigvals,eigvecs = eigen_vals_vecs(Fisher)
  eigvals = eigvals[::-1]
  eigvecs = eigvecs[:,::-1]

  E_vecs = np.real(eigvecs[:,:sd_PCA_size]).T
  S_vecs = np.zeros((sd_PCA_size,Nx_arr))
  for index_pca in range(sd_PCA_size):
    for index_x in range(Nx_arr):
      S_vecs[index_pca][index_x] = np.dot(E_vecs[index_pca],Residual[index_x,:]*delta_ln_z)

  # Create output files
  form = "%.6e" #Output formatting

  # Write file for branching ratio (Evec)
  with open(os.path.join(dir_path,sd_detector_name+"_branching_ratios.dat"),"w") as brfile:
    brfile.write("# In the file there is: z, J_T, J_y, J_mu, E_i (i=1-{})\n".format(sd_PCA_size))
    brfile.write("# The first line contains the number of lines and the number of columns.\n".format(sd_PCA_size))
    brfile.write("{} {}\n".format(Nz_arr,sd_PCA_size))
    for index_z in range(Nz_arr):
      brfile.write((form+" ") % z_arr[index_z])
      brfile.write((form+" ") % f_g[index_z])
      brfile.write((form+" ") % f_y[index_z])
      brfile.write((form    ) % f_mu[index_z])
      for index_pca in range(sd_PCA_size):
        brfile.write((" "+form) % E_vecs[index_pca][index_z])
      brfile.write("\n")

  # Write file for distortion shapes (Svec)
  with open(os.path.join(dir_path,sd_detector_name+"_distortions_shapes.dat"),"w") as dsfile:
    dsfile.write("# In the file there is: nu, G_T, Y_SZ, M_mu, S_i (i=1-{})\n".format(sd_PCA_size))
    dsfile.write("# The first line contains the number of lines and the number of columns.\n".format(sd_PCA_size))
    dsfile.write("{} {}\n".format(Nx_arr,sd_PCA_size))
    for index_x in range(Nx_arr):
      dsfile.write((form+" ") % (x_arr[index_x]*x_to_nu))
      dsfile.write((form+" ") % Gdist[index_x])
      dsfile.write((form+" ") % Ydist[index_x])
      dsfile.write((form    ) % Mdist[index_x])
      for index_pca in range(sd_PCA_size):
        dsfile.write((" "+form) % S_vecs[index_pca][index_x])
      dsfile.write("\n")

  # Update list of detectors
  # Open and read already present list
  with open(os.path.join(dir_path,"detectors_list.dat"),"a") as detector_file:
    if has_noisefile:
      detector_file.write('%s  %s\n' % (sd_detector_name, sd_noisefile_name))
    else:
      detector_file.write('%s  %.6e  %.6e  %.6e  %i  %.6e\n' % (sd_detector_name, sd_detector_nu_min, sd_detector_nu_max, sd_detector_nu_delta, sd_detector_bin_number, sd_detector_delta_Ic))





# Script to change the names of CLASS modules (by Nils Schöneberg & Julien Lesgourgues)
#
# Can be used to:
#  - rename module files, module prefixes, module structures, module structure acronyms
#  - undo renaming
#  - clean the generated log and backup files
#
# usage: CLASS_rename.py [-h] --method {rename,undo,clean} [-v | -q]
#
# optional arguments:
#  -h, --help            show this help message and exit
#  --method {rename,undo,clean}   rename / undo renaming / clean
#  -v, --verbose         Increase the verbosity of the program for more detailed output
#  -q, --quiet           Make the program entirely quiet, setting the verbosity to 0.
#                        Also disables the user confirmation, so use it carefully
#
# The actual renaming to be performed has to be set beforehand in the section below.
# Currently this is set for the transformation
# of CLASS v2.10.8 into CLASS v3.0.0 and backwards.

### EDIT ONLY BELOW ###
### EDIT ONLY BELOW ###

module_filename = ["thermodynamics","perturbations","nonlinear","transfer","spectra"]
module_prefix = ["thermodynamics","perturb","nonlinear","transfer","spectra"]
structure_longname = ["thermo","perturbs","nonlinear","transfers","spectra"]
structure_shortname = ["th","pt","nl","tr","sp"]

newmodule_filename = ["thermodynamics","perturbations","fourier","transfer","harmonic"]
newmodule_prefix = ["thermodynamics","perturbations","fourier","transfer","harmonic"]
newstructure_longname = ["thermodynamics","perturbations","fourier","transfer","harmonic"]
newstructure_shortname= ["th","pt","fo","tr","hr"]

# Potential problem: structure short names are just two
# letters. Combinations of the same two letters may
# appear casually. Thus in some sub-cases we first
# check for exceptions.

# to identify these exception, for each short name (e.g. 'nl'), run:
#
# > grep "nl\." */*.c */*.h */*.py */*.pyx */*.pxd */*.ipynb */*.ini *.ini
#
# > grep "\&nl" */*.c */*.h */*.py */*.pyx */*.pxd */*.ipynb */*.ini *.ini
#
# and check whether some of the lines feature an nl that has nothing
# to do with the stucture short name. If yes, write the exception in
# the dictionary below.

exceptions = {"th":[],
              "pt":[],
              "nl":["nl_corr","R_nl"],
              "tr":[],
              "sp":["osp.","resp"]}

prefix_exceptions = {"thermodynamics":[],
                     "perturbations":[],
                     "nonlinear":["nonlinear_method","nonlinear_scale","nonlinear_min_k_max"],
                     "transfer":[],
                     "spectra":[]}

src_folder = "source"
incl_folder = "include"
test_folder = "test"

### EDIT ONLY ABOVE ###
### EDIT ONLY ABOVE ###




import os
import argparse

# parse the arguments of the command line
parser = argparse.ArgumentParser(description='Change the names of CLASS modules')
parser.add_argument("--method",choices=["rename","undo","clean"], required=True,help="Do you want to rename / undo renaming / clean the generated log and backup files? Type 'rename','undo', or 'clean'")
# default verbosity is 1, can be increased with -v or decreased with -q
group = parser.add_mutually_exclusive_group()
group.add_argument("-v", "--verbose", action="count",default=1,help="Increase the verbosity of the program for more detailed output")
group.add_argument("-q", "--quiet", action="store_true",help="Make the program entirely quiet, setting the verbosity to 0. This also disables the confirmation required by the user, so use it carefully")
parse_dict = parser.parse_args()

if parse_dict.quiet:
  parse_dict.verbose = 0

# Inform the user about the starting of the actual routine
if parse_dict.verbose>0:
  print("START RENAMING ROUTINE v.0.3 (credits Nils Schöneberg & Julien Lesgourgues)")
  print("CHECKING ALL FILES IN DIRECTORY : "+os.path.abspath("."))

# Find the list of all the directories that we will parse and in which we will do changes in some files
#
# Get the list of all folders and subfolders in the local folder
# After this step, each element x is such that x[0] contains a folder name 'folder/subfolder/.../'
folder_list = [x for x in os.walk(".")]
# remove .git, doc, build folders
folder_list = [x for x in folder_list if not (".git" in x[0])]
#folder_list = [x for x in folder_list if not ("doc" in x[0])]
folder_list = [x for x in folder_list if not ("doc/manual" in x[0])]
folder_list = [x for x in folder_list if not ("doc/input/latex" in x[0])]
folder_list = [x for x in folder_list if not ("build" in x[0])]
# remove the folder of the RealSpaceInterface containing cached data
folder_list = [x for x in folder_list if not ("RealSpaceInterface/static" in x[0])]
# keep only the list of all folders, not the files they contain
folder_list = [x[0] for x in folder_list]

if parse_dict.verbose > 0:
  # show the list of 'folder/subfolder/.../'
  print("FOLDER LIST : "+" ".join(folder_list))

  # let the user confirm or abort
  read = input("Continue? (y/n)")
  if not read.startswith("y"):
    quit()

###############
# 'undo' mode #
###############

if parse_dict.method == "undo":

  # For each changed module/file name, go back to old file names (e.g. 'fourier.c' -> 'nonlinear.c') such that they can be overwritten by the corresponding .old files
  for i in range(len(module_filename)):
    xf = module_filename[i]
    yf = newmodule_filename[i]
    os.rename(os.path.join(src_folder,yf+".c"),os.path.join(src_folder,xf+".c"))
    os.rename(os.path.join(incl_folder,yf+".h"),os.path.join(incl_folder,xf+".h"))
    os.rename(os.path.join(test_folder,"test_"+yf+".c"),os.path.join(test_folder,"test_"+xf+".c"))
    if parse_dict.verbose > 0:
      print("REVERTED TO MODULE NAME "+xf)

  # find all folders containing .old and/or .unchanged files
  for fldername in folder_list:
    # First, get name of all files and subfolders in this folder
    filelist_all = os.listdir(fldername)
    filelist = []
    for fname in filelist_all:
      tmp_name = os.path.join(fldername, fname)
      # remove subfolder names, keep only file names
      if os.path.isdir(tmp_name):
        continue
      if tmp_name.endswith(".old"):
        filelist.append(fname)
      elif tmp_name.endswith(".unchanged"):
        filelist.append(fname)

    if parse_dict.verbose > 2:
      print (fldername, filelist)

    for filename in filelist:
      # remove the log files *.unchanged
      # (this can be done safely, as all relevant information is in the .old files)
      if(".unchanged" in filename):
        os.remove(os.path.join(fldername,filename))

      # remove the .old extensions, thus overwriting the changed files with the old files
      if(".old" in filename):
        os.rename(os.path.join(fldername,filename),os.path.join(fldername,filename.replace(".old","")))
        if parse_dict.verbose > 1:
          print ("mv "+os.path.join(fldername,filename)+" "+os.path.join(fldername,filename.replace(".old",""))+"!")

    if parse_dict.verbose > 0:
      print ("IN "+fname+", DELETED .unchanged AND RESTORED ORIGINAL FROM .old FILES")


################
# 'clean' mode #
################

elif parse_dict.method == "clean":

  # find all folders containing .old and/or .unchanged files
  for fldername in folder_list:
    # First, get name of all files and subfolders in this folder
    filelist_all = os.listdir(fldername)
    filelist = []
    for fname in filelist_all:
      tmp_name = os.path.join(fldername, fname)
      # remove subfolder names, keep only file names
      if os.path.isdir(tmp_name):
        continue
      if tmp_name.endswith(".old"):
        filelist.append(fname)
      elif tmp_name.endswith(".unchanged"):
        filelist.append(fname)

    if parse_dict.verbose > 2:
      print (fldername,filelist)

    for filename in filelist:
      # just remove any .unchanged or .old files
      if(".unchanged" in filename or ".old" in filename):
        os.remove(os.path.join(fldername,filename))

    if parse_dict.verbose > 0:
      print ("IN "+fname+", DELETED .unchanged AND .old FILES")

    try:
      # remove log files Makefile.old and possibly autostep.py
      os.remove("Makefile.old")
      if parse_dict.verbose > 0:
        print ("REMOVED Makefile.old")
      os.remove(os.path.join("python","autosetup.py"))
      if parse_dict.verbose > 0:
        print ("REMOVED python/autosetup.py")
    except:
      pass

#################
# 'rename' mode #
#################

elif parse_dict.method == "rename":

  # Some operations only have to be done for the first iteration over all files
  # One example of this is the generation of the backup .old files
  # Thus, we keep track if this is our first iteration
  first_loop = True

  # loop over each module to be renamed/modified
  for i in range(len(module_filename)):

    xf = module_filename[i]
    xp = module_prefix[i]
    xsl = structure_longname[i]
    xss = structure_shortname[i]

    yf = newmodule_filename[i]
    yp = newmodule_prefix[i]
    ysl = newstructure_longname[i]
    yss = newstructure_shortname[i]

    if parse_dict.verbose > 0:
      print("BEGIN RENAMING {} -> {}".format(xsl,ysl))

    # Parse and possibly do changes in each file of the folder fldername
    for fldername in folder_list:

      # Establish the list of file to be parsed and possibly modified in this folder

      # First, get name of all files and subfolders in this folder
      filelist_all = os.listdir(fldername)
      filelist = []
      for fname in filelist_all:
        tmp_name = os.path.join(fldername, fname)
        # remove subfolder names, keep only file names
        if os.path.isdir(tmp_name):
          continue
        # ignore the automatically generated python setup file
        if "autosetup.py" in tmp_name:
          continue
        # take into account other files with extension .c, .py, .pyx, .pxd, .ipynb, .h, .ini
        # but not the .py of the local (root) folder (and thus e.g. not this script!)
        if tmp_name.endswith(".c"):
          filelist.append(fname)
        elif tmp_name.endswith(".py"):
          if fldername != '.':
            filelist.append(fname)
        elif tmp_name.endswith(".pyx"):
          filelist.append(fname)
        elif tmp_name.endswith(".pxd"):
          filelist.append(fname)
        elif tmp_name.endswith(".ipynb"):
          filelist.append(fname)
        elif tmp_name.endswith(".h"):
          filelist.append(fname)
        elif tmp_name.endswith(".ini"):
          filelist.append(fname)
        elif tmp_name.endswith(".md"):
          filelist.append(fname)

      # show the list of file to be parsed and possibly modified in this folder
      if parse_dict.verbose > 1:
        print("WILL MODIFY ALL FILES IN FOLDER '{}': [".format(fldername)+",".join(filelist)+"]")

      # iterate over all files in the current folder
      for filename in filelist:
        # open input file (with old names)
        with open(os.path.join(fldername,filename),"r") as inf:
          # open temporary output file (where we will subsititue the new names)
          with open(os.path.join(fldername,filename+".tmp"),"w") as outf:
            # open a log file with extension .unchanged where we will store lines that were not changed but should have, potentially (for visual inspection)
            with open(os.path.join(fldername,filename+".unchanged"),"a") as unchf:
              # iterate over each line in the input file
              line = inf.readline()
              while line:

                # I. Treat lines where the full structure name appears, e.g. 'nonlinear'
                if "struct "+xsl in line:
                  if "struct "+xsl+" "+xss in line:
                    # replace each structure declaration (e.g. 'struct nonlinear nl' -> 'struct fourier fo')
                    # we isolate this case because it is very useful to catch many occurences of the structure short name (e.g. 'nl') already here
                    line = line.replace("struct "+xsl+" "+xss,"struct "+ysl+" "+yss)
                  else:
                    # replace other occurences (e.g. 'struct nonlinear' -> 'struct fourier')
                    # Special care is needed here! Check that the next character is not a letter
                    # Thus we only allow for a small selection of relevant possibilities
                    for char in ['\t','\n',' ','*','`',';',':']:
                      line = line.replace("struct "+xsl+char,"struct "+ysl+char)
                if "cdef "+xsl in line:
                  if "cdef "+xsl+" "+xss in line:
                    # replace each structure declaration (e.g. 'cdef nonlinear nl' -> 'cdef fourier fo')
                    # we isolate this case because it is very useful to catch many occurences of the structure short name (e.g. 'nl') already here
                    line = line.replace("cdef "+xsl+" "+xss,"cdef "+ysl+" "+yss)
                  else:
                    # replace other occurences (e.g. 'cdef nonlinear' -> 'cdef fourier')
                    line = line.replace("cdef "+xsl,"cdef "+ysl)
                if xsl+" structure" in line:
                  line = line.replace(xsl+" structure",ysl+" structure")

                # II. Treat lines where the module (= file) name appears
                if xf.upper() in line:
                  # replace capitalized module name (e.g. '__NONLINEAR__' -> '__FOURIER__')
                  line = line.replace(xf.upper(),yf.upper())
                if xf+".c" in line:
                  # replace full filename in the comments (e.g. 'nonlinear.c' --> 'fourier.c')
                  line = line.replace(xf+".c",yf+".c")
                if xf+".h" in line:
                  # replace full filename in the comments (e.g. 'nonlinear.h' --> 'fourier.h')
                  line = line.replace(xf+".h",yf+".h")
                if xf+" module" in line:
                  # replace full filename in the comments (e.g. 'nonlinear module' --> 'fourier module')
                  line = line.replace(xf+" module",yf+" module")
                if "\""+xf+"\"" in line:
                  # replace full filename in quotation marks (e.g. '"nonlinear"' --> '"fourier"')
                  line = line.replace("\""+xf+"\"","\""+yf+"\"")


                # III. Treat lines where the prefix appears
                if xp+"_" in line:
                   # replace all prefix names (e.g. 'nonlinear' -> 'fourier')
                   # For all prefix exceptions, substitute the problematic string with 'xx'
                   for i,x in enumerate(prefix_exceptions[xp]):
                     if x in line:
                       line = line.replace(x,prefix_exceptions[xp][i].replace(xp,'xx'))
                   # Now replace all the corresponding names where the prefix appears
                   line = line.replace(xp+"_",yp+"_")
                   # Finally, re-substitute the original exception string instead of the 'xx'
                   for i,x in enumerate(prefix_exceptions[xp]):
                     if x.replace(xp,'xx') in line:
                       line = line.replace(x.replace(xp,'xx'),prefix_exceptions[xp][i])

                # IV. Treat line where short structure name appears, e.g. 'nl'
                if xss in line:

                  # replace pointers towards structure (e.g. 'pnl' -> 'pfo')
                  if "p"+xss in line:
                    line = line.replace("p"+xss,"p"+yss)

                  # replace structure addresses (e.g. '&nl' -> '&fo') and structure short names before dots (e.g. 'nl.error_message' -> 'fo.error_message')
                  if "&"+xss in line or  xss+"." in line:
                    # For all exceptions, substitute the problematic string with 'xx'
                    for i,x in enumerate(exceptions[xss]):
                      if x in line:
                        line = line.replace(x,exceptions[xss][i].replace(xss,'xx'))
                    # Now replace all structure short names before dots and addresses
                    line = line.replace("&"+xss,"&"+yss)
                    line = line.replace(xss+".",yss+".")
                    # Finally, re-substitute the original exception string instead of the 'xx'
                    for i,x in enumerate(exceptions[xss]):
                      if x.replace(xss,'xx') in line:
                        line = line.replace(x.replace(xss,'xx'),exceptions[xss][i])

                  # replace structures as fields of bigger structures in python (e.g. 'self.nl' -> 'self.fo')
                  if "self."+xss in line:
                    line = line.replace("self."+xss,"self."+yss)

                  # if the line did contain the short name in another circumstances, print it in the log file .unchanged
                  if xss in line:
                    # Mark the occurence of the short name by arrows (e.g. 'only' -> 'o-->nl<--y')
                    unchf.write(line.replace(xss,"-->"+xss+"<--"))

                # write the line (changed or not) in the temporary output file
                outf.write(line)
                line = inf.readline()

        # keep the input file but add to it an extension .old (so we keep it as a backup, if something goes wrong) e.g. nonlinear.c -> nonlinear.c.old
        # This is done only in the first loop over modules.
        if first_loop == True:
          os.rename(os.path.join(fldername,filename),os.path.join(fldername,filename+".old"))
        # give to the temporary output file name its final extension (e.g. 'nonlinear.c.tmp' -> 'nonlinear.c')
        os.rename(os.path.join(fldername,filename+".tmp"),os.path.join(fldername,filename))

      if parse_dict.verbose > 1:
        print("SUCCESS IN FOLDER {}".format(fldername))

    # work on the Makefile
    if parse_dict.verbose>1:
      print("MODIFY MAKEFILE")
    with open("Makefile","r") as inf:
      # implement the changes in Makefile.tmp
      with open("Makefile.tmp","w") as outf:
        line = inf.readline()
        while line:
          # replace long names (e.g. 'nonlinear' -> 'fourier')
          if xf in line:
            line = line.replace(xf,yf)
          # replace long names when capitalized
          if xf.upper() in line:
            line = line.replace(xf.upper(),yf.upper())
          outf.write(line)
          line = inf.readline()
    # keep old version with additional .old extension
    if first_loop == True:
      os.rename("Makefile","Makefile.old")
    # rename Makefile.tmp -> Makefile
    os.rename("Makefile.tmp","Makefile")
    if parse_dict.verbose>1:
      print("SUCCESS IN MODIFYING MAKEFILE")

    # change actual file names (e.g. 'nonlinear.c' -> 'fourier.c')
    if parse_dict.verbose>1:
      print("RENAME MODULE "+yf)
    os.rename(os.path.join(src_folder,xf+".c"),os.path.join(src_folder,yf+".c"))
    os.rename(os.path.join(incl_folder,xf+".h"),os.path.join(incl_folder,yf+".h"))
    os.rename(os.path.join(test_folder,"test_"+xf+".c"),os.path.join(test_folder,"test_"+yf+".c"))
    if parse_dict.verbose>1:
      print("SUCCESS IN RENAMING MODULE "+yf)


    if parse_dict.verbose > 0:
      print("SUCCESS FOR RENAMING {} -> {}".format(xf,yf))

    # done for this particular module
    first_loop = False

# end of loop over modulea
if parse_dict.verbose>0:
  print("SUCCESS!")
